#include <sys/types.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <assert.h>
#include <hugetlbfs.h>
#include <sys/stat.h>
#include <sys/syscall.h>
#include <sys/timeb.h>
#include <errno.h>
#include <signal.h>
#include <ctype.h>
#include <termios.h>
#include <sys/time.h>

#include "transcoder.h"

typedef unsigned char u8;
typedef unsigned short u16;
typedef unsigned int u32;
typedef unsigned long long u64;

//#define DEBUG_EN
#ifdef DEBUG_EN
#define pri_dbg(format, ...) printf(format, ##__VA_ARGS__)
#else
#define pri_dbg(format, ...)
#endif

#define QWORD_HI(value) ((value >> 32) & 0xFFFFFFFF)
#define QWORD_LO(value) (value & 0xFFFFFFFF)

#define DIR_EP2RC 0x1
#define DIR_RC2EP 0x2

#define SPECIFIC_EP 0x1
#define SPECIFIC_RC 0x2

#define EN_HUGE 0x1
#define EN_CMP 0x1
#define EN_PERF 0x1

#define DEFAULT_VAL 0x5a
#define DEFAULT_ROUND 1
#define DEFAULT_SIZE 0X1000

#define INVALID_ADDR 0xFF
#define VIRT_ADD 0x1
#define PHY_ADDR 0x2

typedef union {
	u32 flag_riscv;
	struct {
		unsigned flag_riscv0_power_on : 1;
		unsigned flag_riscv0_power_off : 1;
		unsigned flag_riscv0_soft_reset : 1;
		unsigned flag_riscv0_reset_release : 1;
		unsigned flag_riscv0_release_addr : 1;
		unsigned flag_riscv1_power_on : 1;
		unsigned flag_riscv1_power_off : 1;
		unsigned flag_riscv1_soft_reset : 1;
		unsigned flag_riscv1_reset_release : 1;
		unsigned flag_riscv1_release_addr : 1;
		unsigned : 6;
		u16 reserved16;
	} bit;
} flag_riscv_t;

struct t_riscv_control {
	flag_riscv_t riscv_flag;
	u64 specific_pc_addr;
};
struct t_param {
	char *dev_node;
	int direction;
	u64 specific_ep_addr;
	u64 specific_rc_addr;
	u32 tranx_size;
	int en_huge;
	int init_val;
	int init_val_incremental;
	int test_round;
	int compare;
	int perf;
	int out;
	int in;
	char *out_file_name;
	char *in_file_name;
	struct t_riscv_control riscv_control;
};

char *dir[3] = { "null", "ep2rc", "rc2ep" };

static void usage()
{
	printf("\t-d,--dev: device node; e.g. /dev/ama_transcoder0\n");
	printf("\t-r,--dir: ep2rc or rc2ep; default rc2ep\n");
	printf("\t-e,--spe_ep: sep: specific ep address;\n");
	printf("\t-f,--spe_rc: src: specific rc physical address;\n");
	printf("\t-s,--size: transmited data size; default is 0x1000\n");
	printf("\t-g,--huge: use hugepage; default is not\n");
	printf("\t-v,--val: initialize rc memory with this value; default is 0x5a\n");
	printf("\t-a,--add: the data initialized rc memory is incremental; default is disable\n");
	printf("\t-i,--round: test round count; default is 1\n");
	printf("\t-c,--comp: used in rc2ep mode,program will allocate another memory, \n\t\t   then use ep2rc mode readback ep data to it, compare them, default is not\n");
	printf("\t-p,--perf: show hdma transmission performance, not support compare mode, default is disable\n");
	printf("\t-o,--outfile: output ep memory to file; default disable; eg: -o out.bin\n");
	printf("\t-n,--infile: input file needs to be transported to EP memory\n");
	printf("\t-h,--help: display help info;\n");
}

static void set_default_param(struct t_param *param)
{
	param->compare = 0;
	param->perf = 0;
	param->direction = DIR_RC2EP;
	param->en_huge = 0;
	param->init_val = DEFAULT_VAL;
	param->test_round = DEFAULT_ROUND;
	param->tranx_size = DEFAULT_SIZE;
	param->specific_ep_addr = INVALID_ADDR;
	param->specific_rc_addr = INVALID_ADDR;
	param->out = 0;
	param->init_val_incremental = 0;
	param->riscv_control.riscv_flag.flag_riscv = 0;
	param->riscv_control.specific_pc_addr = 0x0;
}

struct option long_options[] = {
	{ "dev", required_argument, NULL, 'd' },
	{ "dir", required_argument, NULL, 'r' },
	{ "spe_ep", required_argument, NULL, 'e' },
	{ "spe_rc", required_argument, NULL, 'f' },
	{ "size", required_argument, NULL, 's' },
	{ "huge", no_argument, NULL, 'g' },
	{ "val", required_argument, NULL, 'v' },
	{ "add", no_argument, NULL, 'a' },
	{ "round", required_argument, NULL, 'i' },
	{ "comp", no_argument, NULL, 'c' },
	{ "perf", no_argument, NULL, 'p' },
	{ "infile", required_argument, NULL, 'n' },
	{ "outfile", required_argument, NULL, 'o' },
	{ "riscv0_power_on", no_argument, NULL, 'A' },
	{ "riscv0_power_off", no_argument, NULL, 'B' },
	{ "riscv0_soft_reset", no_argument, NULL, 'C' },
	{ "riscv0_reset_release", no_argument, NULL, 'D' },
	{ "riscv0_release_addr", required_argument, NULL, 'E' },
	{ "riscv1_power_on", no_argument, NULL, 'F' },
	{ "riscv1_power_off", no_argument, NULL, 'G' },
	{ "riscv1_soft_reset", no_argument, NULL, 'H' },
	{ "riscv1_reset_release", no_argument, NULL, 'I' },
	{ "riscv1_release_addr", required_argument, NULL, 'J' },
	{ "help", no_argument, NULL, 'h' },
	{ 0, 0, 0, 0 }
};
char *const opt_string = "d:r:e:f:s:gv:i:o:achp";

int get_param(int argc, char *argv[], struct t_param *param)
{
	int c;
	int flag[20] = { 0 };
	while ((c = getopt_long_only(argc, argv, opt_string, long_options,
				     NULL)) != -1) {
		switch (c) {
		case 'd': //--dev: device node; e.g. /dev/ama_transcoder0
			if (flag[0] == 1) {
				printf("device param duplication\n");
				return -1;
			}
			flag[0] = 1;
			param->dev_node = optarg;
			pri_dbg("device name is %s.\n", param->dev_node);
			break;
		case 'r': //--dir: ep2rc or rc2ep
			if (flag[1] == 1) {
				pri_dbg("direction param duplication\n");
				return -1;
			}
			flag[1] = 1;
			if (strcmp(optarg, "ep2rc") == 0) {
				param->direction = DIR_EP2RC;
			} else if (strcmp(optarg, "rc2ep") == 0) {
				param->direction = DIR_RC2EP;
			} else {
				printf("direction error\n");
				return -1;
			}
			pri_dbg("tranx direction: %s\n", optarg);
			break;
		case 'e': //--spe_ep: sep: specific ep address
			if (flag[2] == 1) {
				printf("spe_ep param duplication\n");
				return -1;
			}
			flag[2] = 1;
			param->specific_ep_addr = strtol(optarg, NULL, 0);
			pri_dbg("specific ep address: 0x%lx.\n",
				param->specific_ep_addr);
			break;
		case 'f': //--spe_rc: src: specific rc address
			if (flag[3] == 1) {
				printf("spe_rc param duplication\n");
				return -1;
			}
			flag[3] = 1;
			param->specific_rc_addr = strtol(optarg, NULL, 0);
			pri_dbg("specific rc address: 0x%lx.\n",
				param->specific_rc_addr);
			break;
		case 's': //--size: transmited data size
			if (flag[4] == 1) {
				printf("size param duplication\n");
				return -1;
			}
			flag[4] = 1;
			param->tranx_size = strtol(optarg, NULL, 0);
			pri_dbg("data size: 0x%x.\n", param->tranx_size);
			break;
		case 'g': //--huge: use hugepage
			if (flag[5] == 1) {
				printf("huge param duplication\n");
				return -1;
			}
			flag[5] = 1;
			param->en_huge = EN_HUGE;
			pri_dbg("use hugepage: %d.\n", param->en_huge);
			break;
		case 'v': //--val: initialize rc memory with this value
			if (flag[6] == 1) {
				printf("init val param duplication\n");
				return -1;
			}
			flag[6] = 1;
			param->init_val = strtol(optarg, NULL, 0);
			pri_dbg("init value: 0x%x.\n", param->init_val);
			break;
		case 'i': //--round: test round count
			if (flag[7] == 1) {
				printf("round param duplication\n");
				return -1;
			}
			flag[7] = 1;
			param->test_round = strtol(optarg, NULL, 0);
			pri_dbg("round: %d.\n", param->test_round);
			break;
		case 'c': //--comp:compare data
			if (flag[8] == 1) {
				printf("compare param duplication\n");
				return -1;
			}
			flag[8] = 1;
			param->compare = EN_CMP;
			pri_dbg("compare data: %d.\n", param->compare);
			break;
		case 'p': //--perf:show performance
			if (flag[9] == 1) {
				printf("performance param duplication\n");
				return -1;
			}
			flag[9] = 1;
			param->perf = EN_PERF;
			pri_dbg("performance data: %d.\n", param->compare);
			break;
		case 'o': //--outfile, output ep memory to file; e.g. -o out.bin
			if (flag[10] == 1) {
				printf("outfile param duplication\n");
				return -1;
			}
			flag[10] = 1;
			param->out_file_name = optarg;
			param->out = 1;
			pri_dbg("output file name is %s.\n",
				param->out_file_name);
			break;
		case 'n':
			if (flag[12] == 1) {
				printf("infile param duplication\n");
				return -1;
			}
			flag[12] = 1;
			param->in_file_name = optarg;
			param->in = 1;
			pri_dbg("input file name is %s.\n",
				param->in_file_name);
			break;
		case 'a': //--add: increase the initialize data
			if (flag[11] == 1) {
				printf("increase the initialize data param duplication\n");
				return -1;
			}
			flag[11] = 1;
			param->init_val_incremental = 1;
			pri_dbg("incremental: %d.\n",
				param->init_val_incremental);
			break;

		case 'A':
			param->riscv_control.riscv_flag.bit
				.flag_riscv0_power_on = 1;
			break;
		case 'B':
			param->riscv_control.riscv_flag.bit
				.flag_riscv0_power_off = 1;
			break;
		case 'C':
			param->riscv_control.riscv_flag.bit
				.flag_riscv0_soft_reset = 1;
			break;
		case 'D':
			param->riscv_control.riscv_flag.bit
				.flag_riscv0_reset_release = 1;
			break;
		case 'E':
			param->riscv_control.riscv_flag.bit
				.flag_riscv0_release_addr = 1;
			param->riscv_control.specific_pc_addr =
				strtol(optarg, NULL, 0);
			break;
		case 'F':
			param->riscv_control.riscv_flag.bit
				.flag_riscv1_power_on = 1;
			break;
		case 'G':
			param->riscv_control.riscv_flag.bit
				.flag_riscv1_power_off = 1;
			break;
		case 'H':
			param->riscv_control.riscv_flag.bit
				.flag_riscv1_soft_reset = 1;
			break;
		case 'I':
			param->riscv_control.riscv_flag.bit
				.flag_riscv1_reset_release = 1;
			break;
		case 'J':
			param->riscv_control.riscv_flag.bit
				.flag_riscv1_release_addr = 1;
			param->riscv_control.specific_pc_addr =
				strtol(optarg, NULL, 0);
			break;
		case 'h': //--help: display help info
			usage();
			exit(0);

		case '?':
			return -1;
		}
	}
	return 0;
}

int main(int argc, char **argv)
{
	int task_id;
	int fd = -1;
	FILE *output_file = NULL;
	int in_file = -1;
	struct t_param param;

	char *virtual_rc = MAP_FAILED;
	char *comp_addr_rc = MAP_FAILED;
	char *virtual_mmap = MAP_FAILED;
	char *temp_virt = MAP_FAILED;

	struct mem_info ep_region;
	struct release_addr_t release_addr;
	u64 src_addr, dst_addr, compare_addr, ep_addr, rc_addr;
	u32 size;
	int rc_addr_type;
	int direction;

	int i = 0;
	int round, save_round;
	struct trans_pcie_hdma hdma_trans;

	struct timeval time, time_1;
	u64 time_cnt = 0, diff = 0;
	float bw = 0;
	u32 failed_cnt = 1;

	if (argc < 2) {
		usage();
		return 0;
	}

	memset(&param, 0x0, sizeof(param));
	set_default_param(&param);
	if (get_param(argc, argv, &param))
		return -1;

	if (NULL == param.dev_node) {
		usage();
		return -1;
	}
	printf("parmeter: dev_node:%s, direction:%s ep_addr:0x%llx, rc_addr:0x%llx, \n\t  round:%d size:0x%x, init_val:0x%x, en_huge:%d, en_compare:%d\n",
	       param.dev_node, dir[param.direction], param.specific_ep_addr,
	       param.specific_rc_addr, param.test_round, param.tranx_size,
	       param.init_val, param.en_huge, param.compare);
	if ((param.direction == DIR_EP2RC) && (param.compare == EN_CMP)) {
		printf("If transmit data from ep to rc, not support compare data\n");
		return -1;
	}
	if ((param.compare == EN_CMP) && (param.perf == EN_PERF)) {
		printf("performance not support compare mode\n");
		return -1;
	}

	if (param.tranx_size == 0) {
		printf("tranx size = 0x%x\n", param.tranx_size);
		return -1;
	}

	if ((param.direction == DIR_RC2EP) && (param.out == 1)) {
		printf("If transmit data from rc to ep, not support dump memory data to file\n");
		return -1;
	}

	if (param.out == 1) {
		output_file = fopen(param.out_file_name, "wb");
		if (NULL == output_file) {
			printf("failed to open %s.\n", param.out_file_name);
			return -1;
		}
	}

	if (param.in == 1) {
		printf("open in file\n");
		in_file = open(param.in_file_name, O_RDWR);
		if (-1 == in_file) {
			printf("failed to open %s.\n", param.out_file_name);
			goto out_close_infile;
		}
		printf("open in file successfully\n");
	}

	fd = open(param.dev_node, O_RDWR);
	if (fd == -1) {
		printf("Failed to open dev: %s\n", param.dev_node);
		goto out_close_outfile;
	}

	if (param.specific_rc_addr == INVALID_ADDR) {
		if (param.in != 1 && param.out != 1 &&
		    param.riscv_control.riscv_flag.flag_riscv == 0) {
			if (param.en_huge)
				virtual_rc = get_huge_pages(param.tranx_size,
							    GHP_DEFAULT);
			else
				virtual_rc = malloc(param.tranx_size);
			if (!virtual_rc) {
				perror("Err allocate virtual_rc:");
				goto out_close_fd;
			}
			memset(virtual_rc, param.init_val, param.tranx_size);
			rc_addr = (u64)virtual_rc;
			temp_virt = virtual_rc;
			rc_addr_type = VIRT_ADD;
		}
	} else {
		if (param.in != 1 && param.out != 1 &&
		    param.riscv_control.riscv_flag.flag_riscv == 0) {
			rc_addr = param.specific_rc_addr;
			rc_addr_type = PHY_ADDR;

			virtual_mmap = mmap(0, param.tranx_size,
					    PROT_READ | PROT_WRITE, MAP_SHARED,
					    fd, rc_addr);
			if (!virtual_mmap) {
				perror("Err mmap virtual_mmap: ");
				goto out_close_fd;
			}
			memset(virtual_mmap, param.init_val, param.tranx_size);
			temp_virt = virtual_mmap;
		}
	}
	if (param.specific_ep_addr == INVALID_ADDR) {
		if (ioctl(fd, SN_TRANX_MEM_GET_TASKID, &task_id) <
		    0) //get mem task id
		{
			printf("ioctl get task id failed.\n");
			goto out_free_rc;
		}

		ep_region.size = param.tranx_size;
		ep_region.task_id = task_id;
		if (ioctl(fd, SN_TRANX_MEM_ALLOC, &ep_region) < 0) //get ep mem
		{
			printf("ioctl get ep mem failed.\n");
			goto out_free_taskid;
		}
		ep_addr = ep_region.phy_addr;
	} else
		ep_addr = param.specific_ep_addr;

	if (param.compare == EN_CMP) // need link mode to transfer data
	{
		if (param.en_huge)
			comp_addr_rc =
				get_huge_pages(param.tranx_size, GHP_DEFAULT);
		else
			comp_addr_rc = malloc(param.tranx_size);
		if (!comp_addr_rc) {
			perror("Err mmap comp_addr: ");
			goto out_free_epaddr;
		}
		memset(comp_addr_rc, param.init_val + 1, param.tranx_size);
		compare_addr = (u64)comp_addr_rc;
	}

	if (param.direction == DIR_RC2EP) {
		src_addr = rc_addr;
		dst_addr = ep_addr;
		direction = RC2EP;
	} else {
		src_addr = ep_addr;
		dst_addr = rc_addr;
		direction = EP2RC;
	}
	size = param.tranx_size;
	save_round = round = param.test_round;

	printf("hdma_param: dir=%s sar=0x%llx dst=0x%llx size=0x%x round=%d mem_type=%s\n\n",
	       (direction == EP2RC) ? "ep2rc" : "rc2ep", src_addr, dst_addr,
	       size, round,
	       (rc_addr_type == VIRT_ADD) ? "virtual_rc" : "physical_rc");

	while (round) {
		round--;
		if (param.compare != EN_CMP) {
			if (param.in != 1 && param.out != 1 &&
			    param.riscv_control.riscv_flag.flag_riscv == 0) {
				hdma_trans.dar = dst_addr;
				hdma_trans.sar = src_addr;
				hdma_trans.dir = direction;
				hdma_trans.size = size;

				if (param.perf == EN_PERF)
					gettimeofday(&time, NULL);
				if (rc_addr_type == VIRT_ADD) {
					if (ioctl(fd, SN_TRANX_HDMA_TRANX_VIR,
						  &hdma_trans) <
					    0) //support virtual addr
					{
						printf("ioctl SN_TRANX_HDMA_TRANX_VIR failed\n");
						goto out;
					}
				} else {
					if (ioctl(fd, SN_TRANX_HDMA_TRANX_PHY,
						  &hdma_trans) <
					    0) //support physical addr
					{
						printf("ioctl SN_TRANX_HDMA_TRANX_PHY failed\n");
						goto out;
					}
				}

				if ((rc_addr_type == VIRT_ADD) &&
				    (param.out == 1)) {
					fwrite((char *)dst_addr, size, 1,
					       output_file);
				}

				if (param.perf == EN_PERF) {
					gettimeofday(&time_1, NULL);
					diff = 1000000 * (time_1.tv_sec -
							  time.tv_sec) +
					       time_1.tv_usec - time.tv_usec;
					time_cnt += diff;
					//	printf(" round %d/%d time=%lldus.\n", save_round-round, save_round, diff);
				}
			}

			if (param.in == 1) {
				char *infile_virtual_addr = MAP_FAILED;
				size_t filelen = 0;

				filelen = lseek(in_file, 0, SEEK_END);
				lseek(in_file, 0, SEEK_SET);

				rc_addr_type = VIRT_ADD;
				printf("filelen = 0x%lx\n", filelen);
				infile_virtual_addr =
					mmap(NULL, filelen,
					     PROT_READ | PROT_WRITE, MAP_SHARED,
					     in_file, 0);
				if (infile_virtual_addr == MAP_FAILED) {
					perror("mmap failed\n");
					exit(1);
				}

				printf("begin transfer\n");
				hdma_trans.dar = ep_addr;
				hdma_trans.sar = (u64)infile_virtual_addr;
				hdma_trans.dir = RC2EP;
				hdma_trans.size = filelen;

				if (rc_addr_type == VIRT_ADD) {
					if (ioctl(fd, SN_TRANX_HDMA_TRANX_VIR,
						  &hdma_trans) <
					    0) //support virtual addr
					{
						printf("ioctl SN_TRANX_HDMA_TRANX_VIR failed\n");
						goto out;
					}
				}

				munmap(infile_virtual_addr, filelen);
				printf("end transfer\n");
			}

			if (param.out == 1) {
				char *outfile_virtual_addr = NULL;

				printf("param.out == 1\n");
				if (param.en_huge)
					outfile_virtual_addr = get_huge_pages(
						param.tranx_size, GHP_DEFAULT);
				else
					outfile_virtual_addr =
						malloc(param.tranx_size);

				if (!outfile_virtual_addr) {
					printf("alloc memory in host pc failed\n");
					goto out;
				}

				rc_addr_type = VIRT_ADD;

				hdma_trans.dar = (u64)outfile_virtual_addr;
				hdma_trans.sar = ep_addr;
				hdma_trans.dir = EP2RC;
				hdma_trans.size = param.tranx_size;
				printf("param.tranx_size = 0x%x\n",
				       param.tranx_size);

				if (rc_addr_type == VIRT_ADD) {
					if (ioctl(fd, SN_TRANX_HDMA_TRANX_VIR,
						  &hdma_trans) <
					    0) //support virtual addr
					{
						printf("ioctl SN_TRANX_HDMA_TRANX_VIR failed\n");
						goto out;
					}
				}

				fwrite((char *)outfile_virtual_addr,
				       param.tranx_size, 1, output_file);
			}

			if (param.riscv_control.riscv_flag.bit
				    .flag_riscv0_power_on == 1) {
				if (ioctl(fd, RISCV_0_IOCS_POWER_ON, NULL) <
				    0) {
					printf("RISCV_0_IOCS_POWER_ON failed\n");
					goto out;
				}
			}

			if (param.riscv_control.riscv_flag.bit
				    .flag_riscv0_power_off == 1) {
				if (ioctl(fd, RISCV_0_IOCS_POWER_OFF, NULL) <
				    0) {
					printf("RISCV_0_IOCS_POWER_OFF failed\n");
					goto out;
				}
			}

			if (param.riscv_control.riscv_flag.bit
				    .flag_riscv0_soft_reset == 1) {
				if (ioctl(fd, RISCV_0_IOCS_SOFT_RESET, NULL) <
				    0) {
					printf("RISCV_0_IOCS_SOFT_RESET failed\n");
					goto out;
				}
			}

			if (param.riscv_control.riscv_flag.bit
				    .flag_riscv0_reset_release == 1) {
				if (ioctl(fd, RISCV_0_IOCS_RESET_REALEASE,
					  NULL) < 0) {
					printf("RISCV_0_IOCS_RESET_REALEASE failed\n");
					goto out;
				}
			}

			if (param.riscv_control.riscv_flag.bit
				    .flag_riscv0_release_addr == 1) {
				release_addr.pc_l = QWORD_LO(
					param.riscv_control.specific_pc_addr);
				release_addr.pc_h = QWORD_HI(
					param.riscv_control.specific_pc_addr);
				if (ioctl(fd, RISCV_0_IOCS_RELEASE_ADDR,
					  &release_addr) < 0) {
					printf("RISCV_0_IOCS_RELEASE_ADDR failed\n");
					goto out;
				}
			}

			if (param.riscv_control.riscv_flag.bit
				    .flag_riscv1_power_on == 1) {
				if (ioctl(fd, RISCV_1_IOCS_POWER_ON, NULL) <
				    0) {
					printf("RISCV_1_IOCS_POWER_ON failed\n");
					goto out;
				}
			}

			if (param.riscv_control.riscv_flag.bit
				    .flag_riscv1_power_off == 1) {
				if (ioctl(fd, RISCV_1_IOCS_POWER_OFF, NULL) <
				    0) {
					printf("RISCV_1_IOCS_POWER_OFF failed\n");
					goto out;
				}
			}

			if (param.riscv_control.riscv_flag.bit
				    .flag_riscv1_soft_reset == 1) {
				if (ioctl(fd, RISCV_1_IOCS_SOFT_RESET, NULL) <
				    0) {
					printf("RISCV_1_IOCS_SOFT_RESET failed\n");
					goto out;
				}
			}

			if (param.riscv_control.riscv_flag.bit
				    .flag_riscv1_reset_release == 1) {
				if (ioctl(fd, RISCV_1_IOCS_RESET_REALEASE,
					  NULL) < 0) {
					printf("RISCV_1_IOCS_RESET_REALEASE failed\n");
					goto out;
				}
			}

			if (param.riscv_control.riscv_flag.bit
				    .flag_riscv1_release_addr == 1) {
				release_addr.pc_l = QWORD_LO(
					param.riscv_control.specific_pc_addr);
				release_addr.pc_h = QWORD_HI(
					param.riscv_control.specific_pc_addr);
				if (ioctl(fd, RISCV_1_IOCS_RELEASE_ADDR,
					  &release_addr) < 0) {
					printf("RISCV_1_IOCS_RELEASE_ADDR failed\n");
					goto out;
				}
			}

		} else {
			//transmit data from rc to ep
			hdma_trans.dir = RC2EP;
			hdma_trans.sar = rc_addr;
			hdma_trans.dar = ep_addr;
			hdma_trans.size = size;
			if (rc_addr_type == VIRT_ADD) {
				if (ioctl(fd, SN_TRANX_HDMA_TRANX_VIR,
					  &hdma_trans) <
				    0) //support virtual addr
				{
					printf("ioctl SN_TRANX_HDMA_TRANX_VIR RC2EP failed\n");
					goto out;
				}
			} else {
				if (ioctl(fd, SN_TRANX_HDMA_TRANX_PHY,
					  &hdma_trans) <
				    0) //support physical addr
				{
					printf("ioctl SN_TRANX_HDMA_TRANX_PHY RC2EP failed\n");
					goto out;
				}
			}

			//transmit data from ep to rc
			hdma_trans.dir = EP2RC;
			hdma_trans.sar = ep_addr;
			hdma_trans.dar = compare_addr;
			hdma_trans.size = size;
			if (ioctl(fd, SN_TRANX_HDMA_TRANX_VIR, &hdma_trans) <
			    0) //support virtual addr
			{
				printf("ioctl SN_TRANX_HDMA_TRANX_VIR  EP2RC failed\n");
				goto out;
			}

			for (i = 0; i < size; i++) {
				if (*((u8 *)compare_addr + i) !=
				    *((u8 *)temp_virt + i)) {
					printf("%c[40;31m", 0x1b);
					printf("round:%d rc2ep compare failed:%d, total_size:0x%x error_offset:0x%x expected_val:0x%x actual_val:0x%x\n",
					       save_round - round, failed_cnt,
					       size, i, *((u8 *)temp_virt + i),
					       *((u8 *)compare_addr + i));
					printf("%c[0m", 0x1b);
					failed_cnt++;
					break;
				}
			}
#if 1
			if (i == size) {
				printf("round:%d failed_cnt:%d rc2ep compare done, total_size:0x%x, init_val:0x%x\n",
				       save_round - round, failed_cnt - 1, size,
				       *((u8 *)compare_addr + size - 1));
			}
#endif
			if (param.init_val_incremental)
				memset((void *)rc_addr,
				       (save_round - round) % 256, size);
		}
	}

	if (param.perf == EN_PERF) {
		bw = (float)save_round * (float)size * 1000 * 1000 /
		     (float)time_cnt;
		printf("hdma trans size=0x%x, round=%d.\n", size, save_round);
		if (bw > (1024 * 1024))
			printf("total_time=%lld us, avg_bw= %1.2f MB/s\n",
			       time_cnt, bw / 1024 / 1024);
		else if (bw > 1024)
			printf("total_time=%lld us, avg_bw= %1.2f KB/s\n",
			       time_cnt, bw / 1024);
		else
			printf("total_time=%lld us, avg_bw= %1.2f B/s\n",
			       time_cnt, bw);
	}

out:
	if (param.compare == EN_CMP) {
		if (param.en_huge)
			free_huge_pages(comp_addr_rc);
		else
			free(comp_addr_rc);
	}
out_free_epaddr:
	if (param.specific_ep_addr == INVALID_ADDR)
		ioctl(fd, SN_TRANX_MEM_FREE, &ep_region);
out_free_taskid:
	if (param.specific_ep_addr == INVALID_ADDR)
		ioctl(fd, SN_TRANX_MEM_FREE_TASKID, &task_id);
out_free_rc:

	if (param.in != 1 && param.out != 1 &&
	    param.riscv_control.riscv_flag.flag_riscv == 0) {
		if (param.specific_rc_addr == INVALID_ADDR) {
			if (param.en_huge)
				free_huge_pages(virtual_rc);
			else
				free(virtual_rc);
		} else
			munmap(virtual_mmap, param.tranx_size);
	} else {
	}
out_close_fd:
	close(fd);
out_close_infile:
	close(in_file);
out_close_outfile:
	if (param.out == 1)
		fclose(output_file);

	return 0;
}
